%% AMP for linear regression penalized by MCP
% 
% For Gaussian random predictor and response variables
%

close all;

% step of AMP
BP_step = 10000;
% dimension of predictor
M = 100;
% number of predictors
N = 200;
alpha= M/N;

% damping factor
dmp = 0.1;
% number of the set of training samples
n_rand = 1000; 

% MCP parameters
lambda = input('lambda = ');
a = input('a = ');

% choose generative process of response variables
res_ind = input('res_ind(0 or 1) = ');

% density of non-zero components in planted solution
% (used when res_ind = 1)
rho = 0.1;
% st. dev. of non-zero components in planted solution
sigma_x = sqrt(1);
% st. dev. of noise
sigma_y = sqrt(1);

% mean of the response variable
% (used when res_ind = 0)
mu_y_true = 0;

% initialization of output variables
err = zeros(n_rand,1);
df = zeros(n_rand,1);
df_estimator = zeros(n_rand,1);
df_estimator_app = zeros(n_rand,1);

rhoh = zeros(n_rand,1);
pre_err = zeros(n_rand,1);

pre_err_estimator = pre_err;
pre_err_estimator_app = pre_err;

n_r = 0;

% start
while n_r < n_rand
    
    x0 = sigma_x*randn(N,1).*(rand(N,1)<rho);
    A = randn(M,N)/sqrt(M);
    mu = res_ind*A*x0+(1-res_ind)*mu_y_true;
    y = mu+sigma_y*randn(M,1);

    A2 = A.^2;
    X = pinv(A)*y;
    mu_y = A*X;
    v_x = rand(N,1);
    w = y-mu_y;

    shusoku = 0;
    bp_count = 0;

    while (shusoku == 0 && bp_count < BP_step)
    
        bp_count = bp_count + 1;
    
        X_old = X;
        v_x_old = v_x;

        V = A2*v_x;
        w = A*X-V.*((y-w)./(1+V));
        Sigma2 = 1./(A2'*(1./(1+V)));
        R = X + Sigma2.*(A'*((y-w)./(1+V)));

        theta_1 = (abs(R)>lambda*Sigma2).*(abs(R)<=a*lambda);
        theta_2 = (abs(R)>a*lambda);

        var_1 = Sigma2./(1-Sigma2/a);
        var_2 = Sigma2;
        
        % solution of the single-body problem
        X = (R-lambda*Sigma2.*sign(R))./(1-Sigma2/a).*theta_1...
            +R.*theta_2;

        v_x = var_1.*theta_1+var_2.*theta_2;

        shusoku = (sum(abs(X-X_old)<1e-10) == N);
    
        X = dmp*rand(1)*(X-X_old)+X_old;
        v_x = dmp*rand(1)*(v_x-v_x_old)+v_x_old;
        
        mu_y = A*X;
    end

    % memorize quantities
    if bp_count < BP_step
        n_r = n_r +1;

        err(n_r) = (y-mu_y)'*(y-mu_y)/M;
        df(n_r) = (y-mu)'*mu_y/M/sigma_y^2;
        rhoh(n_r) = sum(theta_1+theta_2)/N;
        pre_err(n_r) = err(n_r) + 2*(y-mu)'*mu_y/M;
    
        df_estimator(n_r) = mean(V./(1+V));
        pre_err_estimator(n_r) = err(n_r) + 2*sigma_y^2*df_estimator(n_r);

    end
end
